@@ -1,15 +1,29 @@
+from __future__ import annotations
+
 import io
+from typing import TYPE_CHECKING, Any, cast
 
 import matplotlib.collections as mcollections
 import matplotlib.pyplot as plt
 import numpy as np
 
 from contourpy import FillType, LineType
+from contourpy.util.mpl_util import filled_to_mpl_paths, lines_to_mpl_paths, mpl_codes_to_offsets
+from contourpy.util.renderer import Renderer
 
-from .mpl_util import filled_to_mpl_paths, lines_to_mpl_paths, mpl_codes_to_offsets
+if TYPE_CHECKING:
+    from matplotlib.axes import Axes
+    from matplotlib.figure import Figure
+    from numpy.typing import ArrayLike
+
+    import contourpy._contourpy as cpy
 
 
-class MplRenderer:
+class MplRenderer(Renderer):
+    _axes: Axes
+    _fig: Figure
+    _want_tight: bool
+
     """Utility renderer using Matplotlib to render a grid of plots over the same (x, y) range.
 
     Args:
@@ -23,8 +37,14 @@ class MplRenderer:
             default None.
     """
     def __init__(
-        self, nrows=1, ncols=1, figsize=(9, 9), show_frame=True, backend=None, gridspec_kw=None,
-    ):
+        self,
+        nrows: int = 1,
+        ncols: int = 1,
+        figsize: tuple[int, int] = (9, 9),
+        show_frame: bool = True,
+        backend: str | None = None,
+        gridspec_kw: dict[str, Any] | None = None,
+    ) -> None:
         if backend is not None:
             import matplotlib
             matplotlib.use(backend)
@@ -43,11 +63,11 @@ class MplRenderer:
 
         self._want_tight = True
 
-    def __del__(self):
+    def __del__(self) -> None:
         if hasattr(self, "_fig"):
             plt.close(self._fig)
 
-    def _autoscale(self):
+    def _autoscale(self) -> None:
         # Using axes._need_autoscale attribute if need to autoscale before rendering after adding
         # lines/filled.  Only want to autoscale once per axes regardless of how many lines/filled
         # added.
@@ -58,19 +78,19 @@ class MplRenderer:
         if self._want_tight and len(self._axes) > 1:
             self._fig.tight_layout()
 
-    def _get_ax(self, ax):
+    def _get_ax(self, ax: Axes | int) -> Axes:
         if isinstance(ax, int):
             ax = self._axes[ax]
         return ax
 
-    def _grid_as_2d(self, x, y):
-        x = np.asarray(x)
-        y = np.asarray(y)
-        if x.ndim == 1:
-            x, y = np.meshgrid(x, y)
-        return x, y
-
-    def filled(self, filled, fill_type, ax=0, color="C0", alpha=0.7):
+    def filled(
+        self,
+        filled: cpy.FillReturn,
+        fill_type: FillType,
+        ax: Axes | int = 0,
+        color: str = "C0",
+        alpha: float = 0.7,
+    ) -> None:
         """Plot filled contours on a single Axes.
 
         Args:
@@ -91,7 +111,16 @@ class MplRenderer:
         ax.add_collection(collection)
         ax._need_autoscale = True
 
-    def grid(self, x, y, ax=0, color="black", alpha=0.1, point_color=None, quad_as_tri_alpha=0):
+    def grid(
+        self,
+        x: ArrayLike,
+        y: ArrayLike,
+        ax: Axes | int = 0,
+        color: str = "black",
+        alpha: float = 0.1,
+        point_color: str | None = None,
+        quad_as_tri_alpha: float = 0,
+    ) -> None:
         """Plot quad grid lines on a single Axes.
 
         Args:
@@ -126,10 +155,18 @@ class MplRenderer:
                 np.stack((y[1:, :-1], ymid, y[:-1, 1:])).reshape((3, -1)),
                 **kwargs)
         if point_color is not None:
-            ax.scatter(x, y, color=point_color, alpha=alpha, marker='o')
+            ax.plot(x, y, color=point_color, alpha=alpha, marker='o', lw=0)
         ax._need_autoscale = True
 
-    def lines(self, lines, line_type, ax=0, color="C0", alpha=1.0, linewidth=1):
+    def lines(
+        self,
+        lines: cpy.LineReturn,
+        line_type: LineType,
+        ax: Axes | int = 0,
+        color: str = "C0",
+        alpha: float = 1.0,
+        linewidth: float = 1,
+    ) -> None:
         """Plot contour lines on a single Axes.
 
         Args:
@@ -151,7 +188,14 @@ class MplRenderer:
         ax.add_collection(collection)
         ax._need_autoscale = True
 
-    def mask(self, x, y, z, ax=0, color="black"):
+    def mask(
+        self,
+        x: ArrayLike,
+        y: ArrayLike,
+        z: ArrayLike | np.ma.MaskedArray[Any, Any],
+        ax: Axes | int = 0,
+        color: str = "black",
+    ) -> None:
         """Plot masked out grid points as circles on a single Axes.
 
         Args:
@@ -161,14 +205,14 @@ class MplRenderer:
             ax (int or Matplotlib Axes, optional): Which Axes to plot on, default ``0``.
             color (str, optional): Circle color, default ``"black"``.
         """
-        mask = np.ma.getmask(z)
+        mask = np.ma.getmask(z)  # type: ignore[no-untyped-call]
         if mask is np.ma.nomask:
             return
         ax = self._get_ax(ax)
         x, y = self._grid_as_2d(x, y)
         ax.plot(x[mask], y[mask], 'o', c=color)
 
-    def save(self, filename, transparent=False):
+    def save(self, filename: str, transparent: bool = False) -> None:
         """Save plots to SVG or PNG file.
 
         Args:
@@ -179,7 +223,7 @@ class MplRenderer:
         self._autoscale()
         self._fig.savefig(filename, transparent=transparent)
 
-    def save_to_buffer(self):
+    def save_to_buffer(self) -> io.BytesIO:
         """Save plots to an ``io.BytesIO`` buffer.
 
         Return:
@@ -191,13 +235,13 @@ class MplRenderer:
         buf.seek(0)
         return buf
 
-    def show(self):
+    def show(self) -> None:
         """Show plots in an interactive window, in the usual Matplotlib manner.
         """
         self._autoscale()
         plt.show()
 
-    def title(self, title, ax=0, color=None):
+    def title(self, title: str, ax: Axes | int = 0, color: str | None = None) -> None:
         """Set the title of a single Axes.
 
         Args:
@@ -205,14 +249,24 @@ class MplRenderer:
             ax (int or Matplotlib Axes, optional): Which Axes to set the title of, default ``0``.
             color (str, optional): Color to set title. May be a string color or the letter ``"C"``
                 followed by an integer in the range ``"C0"`` to ``"C9"`` to use a color from the
-                ``tab10`` colormap. Default ``None`` which is ``black``.
+                ``tab10`` colormap. Default is ``None`` which uses Matplotlib's default title color
+                that depends on the stylesheet in use.
         """
         if color:
             self._get_ax(ax).set_title(title, color=color)
         else:
             self._get_ax(ax).set_title(title)
 
-    def z_values(self, x, y, z, ax=0, color="green", fmt=".1f", quad_as_tri=False):
+    def z_values(
+        self,
+        x: ArrayLike,
+        y: ArrayLike,
+        z: ArrayLike,
+        ax: Axes | int = 0,
+        color: str = "green",
+        fmt: str = ".1f",
+        quad_as_tri: bool = False,
+    ) -> None:
         """Show ``z`` values on a single Axes.
 
         Args:
@@ -254,7 +308,7 @@ class MplTestRenderer(MplRenderer):
     No whitespace around plots and no spines/ticks displayed.
     Uses Agg backend, so can only save to file/buffer, cannot call ``show()``.
     """
-    def __init__(self, nrows=1, ncols=1, figsize=(9, 9)):
+    def __init__(self, nrows: int = 1, ncols: int = 1, figsize: tuple[int, int] = (9, 9)) -> None:
         gridspec = {
             "left": 0.01,
             "right": 0.99,
@@ -282,10 +336,24 @@ class MplDebugRenderer(MplRenderer):
     Extends ``MplRenderer`` to add extra information to help in debugging such as markers, arrows,
     text, etc.
     """
-    def __init__(self, nrows=1, ncols=1, figsize=(9, 9), show_frame=True):
+    def __init__(
+        self,
+        nrows: int = 1,
+        ncols: int = 1,
+        figsize: tuple[int, int] = (9, 9),
+        show_frame: bool = True,
+    ) -> None:
         super().__init__(nrows, ncols, figsize, show_frame)
 
-    def _arrow(self, ax, line_start, line_end, color, alpha, arrow_size):
+    def _arrow(
+        self,
+        ax: Axes,
+        line_start: cpy.CoordinateArray,
+        line_end: cpy.CoordinateArray,
+        color: str,
+        alpha: float,
+        arrow_size: float,
+    ) -> None:
         mid = 0.5*(line_start + line_end)
         along = line_end - line_start
         along /= np.sqrt(np.dot(along, along))  # Unit vector.
@@ -297,42 +365,54 @@ class MplDebugRenderer(MplRenderer):
         ))
         ax.plot(arrow[:, 0], arrow[:, 1], "-", c=color, alpha=alpha)
 
-    def filled(self, filled, fill_type, ax=0, color="C1", alpha=0.7, line_color="C0",
-               line_alpha=0.7, point_color="C0", start_point_color="red", arrow_size=0.1):
-        super().filled(filled, fill_type, ax, color, alpha)
-
-        if line_color is None and point_color is None:
-            return
-
-        ax = self._get_ax(ax)
-
+    def _filled_to_lists_of_points_and_offsets(
+        self,
+        filled: cpy.FillReturn,
+        fill_type: FillType,
+    ) -> tuple[list[cpy.PointArray], list[cpy.OffsetArray]]:
         if fill_type == FillType.OuterCode:
+            if TYPE_CHECKING:
+                filled = cast(cpy.FillReturn_OuterCode, filled)
             all_points = filled[0]
             all_offsets = [mpl_codes_to_offsets(codes) for codes in filled[1]]
         elif fill_type == FillType.ChunkCombinedCode:
+            if TYPE_CHECKING:
+                filled = cast(cpy.FillReturn_ChunkCombinedCode, filled)
             all_points = [points for points in filled[0] if points is not None]
             all_offsets = [mpl_codes_to_offsets(codes) for codes in filled[1] if codes is not None]
         elif fill_type == FillType.OuterOffset:
+            if TYPE_CHECKING:
+                filled = cast(cpy.FillReturn_OuterOffset, filled)
             all_points = filled[0]
             all_offsets = filled[1]
         elif fill_type == FillType.ChunkCombinedOffset:
+            if TYPE_CHECKING:
+                filled = cast(cpy.FillReturn_ChunkCombinedOffset, filled)
             all_points = [points for points in filled[0] if points is not None]
             all_offsets = [offsets for offsets in filled[1] if offsets is not None]
         elif fill_type == FillType.ChunkCombinedCodeOffset:
+            if TYPE_CHECKING:
+                filled = cast(cpy.FillReturn_ChunkCombinedCodeOffset, filled)
             all_points = []
             all_offsets = []
             for points, codes, outer_offsets in zip(*filled):
                 if points is None:
                     continue
+                if TYPE_CHECKING:
+                    assert codes is not None and outer_offsets is not None
                 all_points += np.split(points, outer_offsets[1:-1])
-                codes = np.split(codes, outer_offsets[1:-1])
-                all_offsets += [mpl_codes_to_offsets(codes) for codes in codes]
+                all_codes = np.split(codes, outer_offsets[1:-1])
+                all_offsets += [mpl_codes_to_offsets(codes) for codes in all_codes]
         elif fill_type == FillType.ChunkCombinedOffsetOffset:
+            if TYPE_CHECKING:
+                filled = cast(cpy.FillReturn_ChunkCombinedOffsetOffset, filled)
             all_points = []
             all_offsets = []
             for points, offsets, outer_offsets in zip(*filled):
                 if points is None:
                     continue
+                if TYPE_CHECKING:
+                    assert offsets is not None and outer_offsets is not None
                 for i in range(len(outer_offsets)-1):
                     offs = offsets[outer_offsets[i]:outer_offsets[i+1]+1]
                     all_points.append(points[offs[0]:offs[-1]])
@@ -340,6 +420,67 @@ class MplDebugRenderer(MplRenderer):
         else:
             raise RuntimeError(f"Rendering FillType {fill_type} not implemented")
 
+        return all_points, all_offsets
+
+    def _lines_to_list_of_points(
+        self, lines: cpy.LineReturn,
+        line_type: LineType
+    ) -> list[cpy.PointArray]:
+        if line_type == LineType.Separate:
+            if TYPE_CHECKING:
+                lines = cast(cpy.LineReturn_Separate, lines)
+            all_lines = lines
+        elif line_type == LineType.SeparateCode:
+            if TYPE_CHECKING:
+                lines = cast(cpy.LineReturn_SeparateCode, lines)
+            all_lines = lines[0]
+        elif line_type == LineType.ChunkCombinedCode:
+            if TYPE_CHECKING:
+                lines = cast(cpy.LineReturn_ChunkCombinedCode, lines)
+            all_lines = []
+            for points, codes in zip(*lines):
+                if points is not None:
+                    if TYPE_CHECKING:
+                        assert codes is not None
+                    offsets = mpl_codes_to_offsets(codes)
+                    for i in range(len(offsets)-1):
+                        all_lines.append(points[offsets[i]:offsets[i+1]])
+        elif line_type == LineType.ChunkCombinedOffset:
+            if TYPE_CHECKING:
+                lines = cast(cpy.LineReturn_ChunkCombinedOffset, lines)
+            all_lines = []
+            for points, all_offsets in zip(*lines):
+                if points is not None:
+                    if TYPE_CHECKING:
+                        assert all_offsets is not None
+                    for i in range(len(all_offsets)-1):
+                        all_lines.append(points[all_offsets[i]:all_offsets[i+1]])
+        else:
+            raise RuntimeError(f"Rendering LineType {line_type} not implemented")
+
+        return all_lines
+
+    def filled(
+        self,
+        filled: cpy.FillReturn,
+        fill_type: FillType,
+        ax: Axes | int = 0,
+        color: str = "C1",
+        alpha: float = 0.7,
+        line_color: str = "C0",
+        line_alpha: float = 0.7,
+        point_color: str = "C0",
+        start_point_color: str = "red",
+        arrow_size: float = 0.1,
+    ) -> None:
+        super().filled(filled, fill_type, ax, color, alpha)
+
+        if line_color is None and point_color is None:
+            return
+
+        ax = self._get_ax(ax)
+        all_points, all_offsets = self._filled_to_lists_of_points_and_offsets(filled, fill_type)
+
         # Lines.
         if line_color is not None:
             for points, offsets in zip(all_points, all_offsets):
@@ -367,34 +508,25 @@ class MplDebugRenderer(MplRenderer):
                     ax.plot(points[:, 0][start_indices], points[:, 1][start_indices], "o",
                             c=start_point_color, alpha=line_alpha)
 
-    def lines(self, lines, line_type, ax=0, color="C0", alpha=1.0, linewidth=1, point_color="C0",
-              start_point_color="red", arrow_size=0.1):
+    def lines(
+        self,
+        lines: cpy.LineReturn,
+        line_type: LineType,
+        ax: Axes | int = 0,
+        color: str = "C0",
+        alpha: float = 1.0,
+        linewidth: float = 1,
+        point_color: str = "C0",
+        start_point_color: str = "red",
+        arrow_size: float = 0.1,
+    ) -> None:
         super().lines(lines, line_type, ax, color, alpha, linewidth)
 
         if arrow_size == 0.0 and point_color is None:
             return
 
         ax = self._get_ax(ax)
-
-        if line_type == LineType.Separate:
-            all_lines = lines
-        elif line_type == LineType.SeparateCode:
-            all_lines = lines[0]
-        elif line_type == LineType.ChunkCombinedCode:
-            all_lines = []
-            for points, codes in zip(*lines):
-                if points is not None:
-                    offsets = mpl_codes_to_offsets(codes)
-                    for i in range(len(offsets)-1):
-                        all_lines.append(points[offsets[i]:offsets[i+1]])
-        elif line_type == LineType.ChunkCombinedOffset:
-            all_lines = []
-            for points, offsets in zip(*lines):
-                if points is not None:
-                    for i in range(len(offsets)-1):
-                        all_lines.append(points[offsets[i]:offsets[i+1]])
-        else:
-            raise RuntimeError(f"Rendering LineType {line_type} not implemented")
+        all_lines = self._lines_to_list_of_points(lines, line_type)
 
         if arrow_size > 0.0:
             for line in all_lines:
@@ -413,7 +545,14 @@ class MplDebugRenderer(MplRenderer):
                 ax.plot(line[start_index:end_index, 0], line[start_index:end_index, 1], "o",
                         c=color, alpha=alpha)
 
-    def point_numbers(self, x, y, z, ax=0, color="red"):
+    def point_numbers(
+        self,
+        x: ArrayLike,
+        y: ArrayLike,
+        z: ArrayLike,
+        ax: Axes | int = 0,
+        color: str = "red",
+    ) -> None:
         ax = self._get_ax(ax)
         x, y = self._grid_as_2d(x, y)
         z = np.asarray(z)
@@ -424,7 +563,14 @@ class MplDebugRenderer(MplRenderer):
                 ax.text(x[j, i], y[j, i], str(quad), ha="right", va="top", color=color,
                         clip_on=True)
 
-    def quad_numbers(self, x, y, z, ax=0, color="blue"):
+    def quad_numbers(
+        self,
+        x: ArrayLike,
+        y: ArrayLike,
+        z: ArrayLike,
+        ax: Axes | int = 0,
+        color: str = "blue",
+    ) -> None:
         ax = self._get_ax(ax)
         x, y = self._grid_as_2d(x, y)
         z = np.asarray(z)
@@ -436,7 +582,16 @@ class MplDebugRenderer(MplRenderer):
                 ymid = y[j-1:j+1, i-1:i+1].mean()
                 ax.text(xmid, ymid, str(quad), ha="center", va="center", color=color, clip_on=True)
 
-    def z_levels(self, x, y, z, lower_level, upper_level=None, ax=0, color="green"):
+    def z_levels(
+        self,
+        x: ArrayLike,
+        y: ArrayLike,
+        z: ArrayLike,
+        lower_level: float,
+        upper_level: float | None = None,
+        ax: Axes | int = 0,
+        color: str = "green",
+    ) -> None:
         ax = self._get_ax(ax)
         x, y = self._grid_as_2d(x, y)
         z = np.asarray(z)
